/*
	Name: usb.h
	Date: 16.Dec.2012
	Encoding: UTF-8
	Author: Michal Riša <xrisam01@stud.fit.vutbr.cz>
	Licence:
		Copyright (C) 2013 Michal Riša <mh74s74@gmail.com>

		This file is part of bp_xrisam01.

		bp_xrisam01 is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License version 3 as
		published by the Free Software Foundation.

		bp_xrisam01 is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
		GNU General Public License for more details.

		You should have received a copy of the GNU General Public License
		along with bp_xrisam01. If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef USB_H
#define USB_H

#include "types.h"
#include "usb_drv.h"

/* Standardne USB datove struktury */

//PIDs
enum {
	/*            xx3210xx */
	/* token */
	PID_OUT   = 0b00000100,
	PID_IN    = 0b00100100,
	PID_SOF   = 0b00010100,
	PID_SETUP = 0b00110100,
	/* data */
	PID_DATA0 = 0b00001100,
	PID_DATA1 = 0b00101100,
	PID_DATA2 = 0b00011100,
	PID_MDATA = 0b00111100,
	/* handshake */
	PID_ACK   = 0b00001000,
	PID_NAK   = 0b00101000,
	PID_STALL = 0b00111000,
	PID_NYET  = 0b00011000,
	/* special */
//	PID_SPEC  = 0b00110000, //? 0000
	PID_ERR   = 0b00110000,
	PID_SPLIT = 0b00100000,
	PID_PING  = 0b00010000,
//	reserved  = 0b00000000
};

//------------------------------------------------------------------------------
//format setup transakcie a hodnoty jej poloziek
struct usb_setup_pkt {
	uint8_t bmRequestType;
	uint8_t bRequest;
	uint16_t wValue;
	uint16_t wIndex;
	uint16_t wLength;
};

struct usb_setup_pkt_bytes { //viacbajtove polozky pristupne po bajtoch
	uint8_t bmRequestType;
	uint8_t bRequest;
	uint8_t wValue[2];
	uint8_t wIndex[2];
	uint8_t wLength[2];
};

union ep0out_u {
	uint8_t buf[USB_EP0_BUF_SIZE];
	struct usb_setup_pkt pkt;
	struct usb_setup_pkt_bytes bytes;
	struct usb_setup_pkt_std {
		uint8_t bmRequestType;
		uint8_t bRequest;
		uint16_t wValue;
		uint16_t wIndex;
		uint16_t wLength;
	}std;
	struct {
		//bmRequestType
		unsigned Recipient:5;
		unsigned Type:2;
		unsigned Direction:1;
		//bRequest
		unsigned:8;
		//wValue
		unsigned:8;
		unsigned:8;
		//wIndex
		unsigned:8;
		unsigned:8;
		//wLength
		unsigned:8;
		unsigned:8;
	}bmRequestType;
	struct {
		//bmRequestType
		unsigned:8;
		//bRequest
		unsigned:8;
		//wValue
		unsigned:8;
		unsigned:8;
		//wIndex
		unsigned Endpoint:4;
		unsigned:3;
		unsigned Direction:1;
		unsigned:8;
		//wLength
		unsigned:8;
		unsigned:8;
	}wIndex_ep;
	/* wIndex_interface === bytes.wIndex[0] */
};

//usb_setup_pkt.bRequest standardne prikazy (standard device requests)
enum {
	GET_STATUS = 0,
	CLEAR_FEATURE = 1,
	SET_FEATURE = 3,
	SET_ADDRESS = 5,
	GET_DESCRIPTOR = 6,
	SET_DESCRIPTOR = 7,
	GET_CONFIGURATION = 8,
	SET_CONFIGURATION = 9,
	GET_INTERFACE = 10,
	SET_INTERFACE = 11,
	SYNCHR_FRAME = 12
};

/*
usb_setup_pkt.bmRequestType:

--data direction--
7:  0 = host to device (out)
    1 = device to host (in)
--type--
6:5 0 = standard
    1 = class
    2 = vendor
    3 = reserved
--recipient--
4:0 0 = device
    1 = interface
    2 = endpoint
    3 = other
    4...31 = reserved
*/

//std_bmRequestType.Recipient, Type, Direction
enum {
	DIR_HOST_TO_DEV = 0,
	DIR_DEV_TO_HOST = 1,
	TYPE_STD = 0,
	TYPE_CLASS = 1,
	TYPE_VENDOR = 2,
	RCPT_DEVICE = 0,
	RCPT_INTERFACE = 1,
	RCPT_ENDPOINT = 2,
	/* feature selektory */
	FT_DEVICE_REMOTE_WAKEUP = 1,
	FT_ENDPOINT_HALT = 0,
	FT_TEST_MODE = 2,
};

//std_wValue.h pre descriptor type
enum { 
	DESC_DEV = 1,
	DESC_CFG = 2,
	DESC_STR = 3,
	DESC_IFACE = 4,
	DESC_EP = 5,
	DESC_QUALIFIER = 6,
	DESC_OTHER_SPEED_CFG = 7,
	DESC_IFACE_POWER = 8,
};

//endpoint transfer type
enum { /*  */
	EPT_CTRL = 0, /* CONTROL */
	EPT_ISO  = 1, /* ISOCHRONOUS */
	EPT_BULK = 2, /* BULK */
	EPT_INT  = 3, /* INTERRUPT */
};
//------------------------------------------------------------------------------
//deskriptory
struct usb_device_desc { /* device descriptor, 18B */
	uint8_t bLength;
	uint8_t bDescriptorType;
	uint16_t bcdUSB;
	uint8_t bDeviceClass;
	uint8_t bDeviceSubClass;
	uint8_t bDeviceProtocol;
	uint8_t bMaxPacketSize;
	uint16_t idVendor;
	uint16_t idProduct;
	uint16_t bcdDevice;
	uint8_t iManufacturer;
	uint8_t iProduct;
	uint8_t iSerialNumber;
	uint8_t bNumConfigurations;
};

struct usb_config_desc { /* config descriptor, 9B */
	uint8_t bLength;
	uint8_t bDescriptorType;
	uint16_t wTotalLength;
	uint8_t bNumInterfaces;
	uint8_t bConfigurationValue;
	uint8_t iConfiguration;
	uint8_t bmAttributes;
	uint8_t bMaxPower;
};

struct usb_interface_desc { /* interface descriptor, 9B */
	uint8_t bLength;
	uint8_t bDescriptorType;
	uint8_t bInterfaceNumber;
	uint8_t bAlternateSetting;
	uint8_t bNumEndpoints;
	uint8_t bInterfaceClass;
	uint8_t bInterfaceSubClass;
	uint8_t bInterfaceProtocol;
	uint8_t iInterface;
};

struct usb_endpoint_desc { /* endpoint descriptor, 7B */
	uint8_t bLength;
	uint8_t bDescriptorType;
	uint8_t bEndpointAddress;
	uint8_t bmAttributes;
	uint16_t wMaxPacketSize;
	uint8_t bInterval;
};

struct usb_device_qualifier { /* device qualifier */
	uint8_t bLength;
	uint8_t bDescriptorType;
	uint16_t bcdUSB;
	uint8_t bDeviceClass;
	uint8_t bDeviceSubClass;
	uint8_t bDeviceProtocol;
	uint8_t bMaxPacketSize0;
	uint8_t bNumConfigurations;
	uint8_t bReserved;
};
//------------------------------------------------------------------------------
//remote wakeup a self power priznaky
union usb_status_t {
	struct {
		unsigned:6;
		unsigned Remote_wakeup:1;
		unsigned Self_powered:1;
	}bits;
	uint8_t bytes[1];
};

#endif
