/*
	Name: usb.c
	Date: 15.Jan.2013
	Encoding: UTF-8
	Author: Michal Riša <xrisam01@stud.fit.vutbr.cz>
	Licence:
		Copyright (C) 2013 Michal Riša <mh74s74@gmail.com>

		This file is part of bp_xrisam01.

		bp_xrisam01 is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License version 3 as
		published by the Free Software Foundation.

		bp_xrisam01 is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
		GNU General Public License for more details.

		You should have received a copy of the GNU General Public License
		along with bp_xrisam01. If not, see <http://www.gnu.org/licenses/>.
*/

#include <hidef.h>
#include "derivative.h"

#include "usb_drv.h"
#include "usb.h"

//------------------------------------------------------------------------------
//prototypy
static void usb_OnReset(void);
static void usb_suspend(void);
//------------------------------------------------------------------------------
//globalne premenne
struct usb_bdt bdt @0x1860; //buffer descriptor table v usb ram
uint8_t usb_state;          //stav usb zariadenia
void (*ep_hdrs[7])(void);   //obsluzne fcie pre transakcie endpointov
uint8_t cur_cfg;            //aktivna konfiguracia zariadenia, 0 = ziadna

union usb_status_t usb_status; /* GET_STATUS */

/*
Funkcia: usb_init
Popis: Zapne USB modul a inicializuje USB ovladac.
*/
void usb_init(void) {
#ifdef USB_REMOTE_WAKEUP
	usb_status.bits.Remote_wakeup = 1;
#endif

#ifdef USB_SELF_POWERED
	usb_status.bits.Self_powered = 1;
#endif

	//reset USB modulu
	USBCTL0_USBRESET = 1;
	usb_state = USB_ATTACHED;

#ifdef VREG_EN
	//pull-up, vreg, transciever
	USBCTL0 = 0b01000101;
#else
	//pull-up, transciever
	USBCTL0 = 0b01000001;
#endif

	//zapne USB modul
	CTL_USBEN = 1;
	//povoli USB bus reset prerusenie
	INTENB = 1;
}
//------------------------------------------------------------------------------
//hodiny

/*
Funkcia: usb_clock_init
Popis: Nastavi hodinovy system MC9S08JM60 pre pouzitie USB modulu.
Zdroj: Freescale 3560,
       "Example 1. MCG Initialization Routine Using External 12 MHz Crystal"
*/
void usb_clock_init(void) {
	/* the MCG is default set to FEI mode, it should be change to FBE mode*/
	MCGC2 = 0x36;
	while(!(MCGSC & 0x02)); /* wait for the OSC stable */
	MCGC1 = 0x9B;
	while((MCGSC & 0x1C) != 0x08); /* check the external reference clock is
		selected or not */
	MCGC3 = 0x48; /* switch to PBE mode from FBE */
	while ((MCGSC & 0x48) != 0x48); /* wait for the PLL is locked & */
	MCGC1 &= 0x3F; /* Switch to PEE mode from PBE mode */
	while((MCGSC & 0x6C) != 0x6C);
}
//------------------------------------------------------------------------------
//obsluha udalosti
/*
Funkcia: usb_OnReset
Popis: Spracuje USB bus reset signal.
*/
static void usb_OnReset(void) {
	if(cur_cfg != 0)
		/* deaktivuje beziacu konfiguraciu */
		cfg_hdrs[cur_cfg - 1](EVENT_DEACTIVATE);
	cur_cfg = 0; /* ziadna konfiguracia */

	ADDR = 0; /* zakladna adresa */
	INTENB = 0;
	ERRENB = 0;

	bdt.ep0in.epadr = EPADR(ep0in_buf); /* adresy do bdt */
	bdt.ep0out.epadr = EPADR(ep0out_setup);
	ep0_next_trf(); /* pripravi dalsi prenos */
	EPCTL1 = 0;
	EPCTL2 = 0;
	EPCTL3 = 0;
	EPCTL4 = 0;
	EPCTL5 = 0;
	EPCTL6 = 0;

	while(INTSTAT_TOKDNEF) /* splachne SIE FIFO */
		INTSTAT = 0xBF;

	EPCTL0 = EP_RX | EP_TX | EP_HSK; /* zapne ep0 */

	/* odstrani priznaky a povoli prerusenia */
	ERRSTAT = 0xFF;
	ERRENB = 0xFF;
	INTSTAT = 0xFF;
	INTENB = 0x9F; /* vsetko okrem RESUMEF */
	usb_state = USB_DEFAULT;
}

/*
Funkcia: check_remote_wakeup
Navratova hodnota: 1 = nastali podmienky pre remote wakeup, 0 = nenastali.
Popis: Kontrola nastania podmienok pre remote wakeup. Ak nastali, posle hostovi
       RESUME signalizaciu.
*/
uint8_t check_remote_wakeup(void) {
	static uint16_t delay = 8000;

	if(usb_status.bits.Remote_wakeup == 0)
		//remote wakeup je vypnuty
		return 0;

	if(usb_CheckRemoteWakeup() == 0)
		//mcu neprebral remote wakeup
		return 0;

	//spusti RESUME signalizaciu po dobu 1 az 15 ms
	CTL_CRESUME = 1;
	while(delay) {
		delay--;
	}
	CTL_CRESUME = 0;
	return 1;
}

/*
Funkcia: usb_suspend
Popis: Funkcia pre spracovanie SLEEPF prerusenia. Self powered zariadnie neide
       do stop3 rezimu. Bus powered zariadenie prejde do stop3 rezimu. Funkcia
       aplikacii do obsluznej funkcie udalosti konfiguracie posle prislusne
       udalosti.

       Bus-powered zariadenie s povolenym remote wakeup po prebrati zo stop3
       rezimu moze v obsluznej funckii usb_OnCheckRemoteWakeup vratit hodnotu 1
       cim spusti RESUME signalizaciu.
*/
static void usb_suspend(void) {
	uint8_t event = 255;

	if(cur_cfg != 0)
		//oznami aktivnej konfiguracii suspend udalost
		cfg_hdrs[cur_cfg - 1](EVENT_SUSPEND);

	if(usb_status.bits.Self_powered) {
		//zariadenie ma vlastne napajanie, neide do stop3 rezimu
		INTENB_RESUME = 1;
		INTENB_SLEEP = 0;
		return;
	}

	//zariadenie je bus-powered, prejde do stop3 rezimu
	for(;;) {
		uint8_t delay = 100;
		//zakaze vsetky USB prerusenia
		INTENB = 0;
		//povoli low power resume event
		USBCTL0_USBRESMEN = 1;

//		PTFDD_PTFDD1 = 1; /* output */

		//prepne MCU do stop3 rezimu
		asm stop;

//		PTFD_PTFD1 = !PTFD_PTFD1;

		//pocka na odoznenie prechodovych javov
		while(delay) {
			delay--;
		}

		if(check_remote_wakeup()) {
			//prebudenie sposobila udalost spustajuca remote wakeup
			event = EVENT_WAKEUP;
			break;
		}

		if(INTSTAT_RESUMEF == 0 && INTSTAT_USBRSTF == 0)
			//znovu ide do stop3 rezimu
			continue;

		if(INTSTAT_RESUMEF) {
			//prebudenie sposobil URB resume signal 
			INTSTAT_RESUMEF = 1;
			event = EVENT_WAKEUP;
		}

		if(INTSTAT_USBRSTF) {
			/* prebudenie sposobil USB bus reset signal, reset neposiela
				udalost beziacej konfiguracii => event = 255 */

			/* aktivnej konfiguracii posle EVENT_DEACTIVATE a nastavi USB modul
			   do zakladnych nastaveni */
			usb_OnReset();
			INTSTAT_USBRSTF = 1;
		}
		break;
	}

	if(cur_cfg != 0 && event != 255)
		//beziacej konfiguracii posle udalost konfiguracie
		cfg_hdrs[cur_cfg - 1](event);
	INTENB = 0x9F;
}

/*
Funkcia: usb_isr
Popis: Obsluzna funkcia preruseni USB modulu.
*/
__interrupt VectorNumber_Vusb void usb_isr(void) {
	static uint8_t suspended = 0;
	if(USBCTL0_LPRESF) {
		//vypne low power resume event
		USBCTL0_USBRESMEN = 0;
		return;
	}
	if(INTSTAT_RESUMEF && INTENB_RESUME) {
		//self-powered only
		//prepne USB modul do normalneho rezimu
		INTENB_RESUME = 0;
		INTENB_SLEEP = 1;
		INTSTAT_RESUMEF = 1;
		CTL_TSUSPEND = 0;
		if(cur_cfg != 0)
			//prebudi beziacu konfiguraciu
			cfg_hdrs[cur_cfg - 1](EVENT_WAKEUP);
		return;
	}
	if(INTSTAT_SLEEPF && INTENB_SLEEP) {
		//neaktivita zbernice po t > 3ms
		INTSTAT_SLEEPF = 1;
		//zabrani pripadnemu vnaraniu ISR, nieco ako kriticka sekcia
		if(suspended == 0) {
			suspended = 1;
			usb_suspend();
			suspended = 0;
		}
		return;
	}
	if(INTSTAT_USBRSTF && INTENB_USBRST) {
		//USB bus reset signal
		usb_OnReset();
		INTSTAT_USBRSTF = 1;
	}
	if(INTSTAT_ERRORF && INTENB_ERROR) {
		//chyba
		if((STAT & 0xF0) != 0x00)
			//zavola obsluznu funkciu aplikacie
			usb_OnError();
		ERRSTAT = 0xFF;
		INTSTAT_ERRORF = 1;
	}
	if(INTSTAT_TOKDNEF && INTENB_TOKDNE) {
		//dokoncena transakcia
		uint8_t ep = (STAT & 0xF0) >> 4;
		if(ep == 0)
			//obsluha endpointu 0 je radsej mimo obsluznych funckii
			ep0_hdr();
		else
			//zavola obsluznu funkciu endpointu v aplikacii
			ep_hdrs[ep]();
		INTSTAT_TOKDNEF = 1;
	}
	if(INTSTAT_STALLF && INTENB_STALL) {
		//poslany stall handshake
		uint8_t *reg;
		uint8_t cnt;

		//cislo endpointu v STAT je nepouzitelne aj ked datasheet uvadza inak
		if(EPCTL0_EPSTALL == 1) {
			//stall poslany na endpointe 0
			EPCTL0_EPSTALL = 0;
			ep0_next_trf();
		}

		//najde endpoint ktory poslal STALL
		for(reg = &EPCTL1, cnt = 1; cnt < 7; cnt++, reg++) {
			/* if(EPCTL(cnt)_EPSTALL) { */
			if(*reg & 0x2)
				//zavola obsluznu funkciu aplikacie s cislom STALL-ed endpointu
				usb_OnStallSent(cnt);
		}

		INTSTAT_STALLF = 1;
	}
	if(INTSTAT_SOFTOKF && INTENB_SOFTOK) {
		//SOF paket prijaty
		usb_OnSof();
		INTSTAT_SOFTOKF = 1;
	}
}
