/*
	Name: usb_ep0.c
	Date: 1.Mar.2013
	Encoding: UTF-8
	Author: Michal Riša <xrisam01@stud.fit.vutbr.cz>
	Licence:
		Copyright (C) 2013 Michal Riša <mh74s74@gmail.com>

		This file is part of bp_xrisam01.

		bp_xrisam01 is free software: you can redistribute it and/or modify
		it under the terms of the GNU General Public License version 3 as
		published by the Free Software Foundation.

		bp_xrisam01 is distributed in the hope that it will be useful,
		but WITHOUT ANY WARRANTY; without even the implied warranty of
		MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
		GNU General Public License for more details.

		You should have received a copy of the GNU General Public License
		along with bp_xrisam01. If not, see <http://www.gnu.org/licenses/>.
*/

#include <hidef.h>

#include "derivative.h"
#include "usb.h"
#include "usb_drv.h"

//deskriptory
#include "usb_descriptors.h"
//------------------------------------------------------------------------------
//prototypy a globalne premenne
void ep0_hdr(void);
void ep0_next_trf(void);

static uint8_t ep0_tx_data(uint8_t *data, uint16_t bytes);
static void ep0_setup(void);
static void ep0_setup_complete(void);
static uint8_t get_descriptor_hdr(void);
static uint8_t set_configuration_hdr(void);
static uint8_t get_configuracion_hdr(void);
static uint8_t feature_hdr(uint8_t set);
static uint8_t get_status_hdr(void);
static uint8_t get_interface_hdr(void);
static uint8_t set_interface_hdr(void);

//buffre endpointu 0
union ep0out_u ep0out_setup               @0x1880;
unsigned char ep0in_buf[USB_EP0_BUF_SIZE] @0x1890;

//stavy riadiaceho prenosu
enum ctrl_trf_state {
//setup faza --------------------------------------
	CTRL_TRF_SETUP,      //prijem setup transakcie
//datova faza -------------------------------------
	CTRL_TRF_DATA_IN,    //control read
	CTRL_TRF_DATA_OUT,   //control write
	CTRL_TRF_TRUNC,      //skratenie control read
//statova faza ------------------------------------
	CTRL_TRF_STATUS_IN,  //status pre control write
	CTRL_TRF_STATUS_OUT, //status pre control read
};

struct ctrl_trf ctrl;
//------------------------------------------------------------------------------
//endpointy

/*
Funkcia: ep0_tx_data
Parametre:
	data:
		NULL: ocakava data uz vlozene do ep0in_buf
		inak: z data vlozi tolko bajtov dat kolko sa naraz zmesti do ep0in_buf
	bytes: velkost dat na poslanie
Navratova hodnota:
	data == NULL: velkost ep0in_buf v bajtoch.
	data != NULL: mnozstvo dat vlozenych do ep0in_buf.
Popis: ^.
*/
static uint8_t ep0_tx_data(uint8_t *data, uint16_t bytes) {
	//vyberie tolko dat kolko sa zmesti do buffra endpointu
	uint8_t limit;
	if(bytes > sizeof ep0in_buf)
		limit = sizeof ep0in_buf;
	else
		limit = (uint8_t)bytes;

	//vlozi data do buffra endpointu ak tam este niesu (NULL)
	if(data != NULL) {
		uint8_t pos;
		for(pos = 0; pos < limit; pos++, data++)
			ep0in_buf[pos] = *data;
	}

	//posle
	bdt.ep0in.bc = limit;
	if(bdt.ep0in.u.bits.DATA)
		bdt.ep0in.u.reg = EPF_OWN_SIE | EPF_DTS | EPF_DATA1;
	else
		bdt.ep0in.u.reg = EPF_OWN_SIE | EPF_DTS | EPF_DATA0;

	return limit;
}

/*
Funkcia: ep0_send_data
Popis: Posle data cez control read prenos, nastavi konecny automat riadiaceho
       prenosu a riesi kratsie prenosy ako host ziadal.
*/
static void ep0_send_data(void) {
	uint8_t done = ep0_tx_data(ctrl.data, ctrl.bytes);
	ctrl.bytes     -= done;
	ctrl.data      += done;

	if(ctrl.bytes == 0) {
		//zariadenie poslalo vsetko co chcelo
		if(ctrl.trunc)
			/* host chcel aby zariadenie poslalo viac a zariadenie poslalo
			   nasobok velkosti buffra, zariadenie posle kratky paket ktory
			   hostovi naznaci koniec prenosu */
			ctrl.state = CTRL_TRF_TRUNC;
		else
			//nasleduje stavova faza opacneho smeru: prenos IN, stav OUT
			ctrl.state = CTRL_TRF_STATUS_OUT;
	}
}

#if 0
/*
Funkcia:
Parametre:
Navratova hodnota:
Popis:
*/
static void ep0_rx(void) {
	/* prijimajuci endpoint neguje data toggle bit po prijati
		transakcie */
	bdt.ep0out.u.bits.DATA = !bdt.ep0out.u.bits.DATA;
}
#endif
//------------------------------------------------------------------------------
//spracovanie transakcii a prenosov na control pipe

/*
riadiaci prenos:

druh    | setup      | data                                | status
prenosu | transakcia | transakcie                          | transakcia
-----------------------------------------------------------------------
read    | OUT/DATA0  | IN/DATA1  IN/DATA0  ... IN/DATA0/1  | OUT/DATA1
write   | OUT/DATA0  | OUT/DATA1 OUT/DATA0 ... OUT/DATA0/1 | IN/DATA1
-----------------------------------------------------------------------
no-data | OUT/DATA0  | [vynechane]                         | IN/DATA1
*/

/*
Funkcia: ep0_hdr
Popis: Spracuje udalosti endpointov 0 IN a 0 OUT. Implementuje konecny automat
       riadiaceho prenosu.
*/
void ep0_hdr(void) { /* TOKDNEF */
	if(STAT_IN) {
		/* transakcia poslana */
		/* vysielajuci endpoint neguje data toggle bit po spravnom poslani
		   transakcie */
		bdt.ep0in.u.bits.DATA = !bdt.ep0in.u.bits.DATA;
		switch(ctrl.state) {
			case CTRL_TRF_TRUNC:
				//posle kratky paket na naznacenie konca prenosu
				(void)ep0_tx_data(NULL, 0);
				ctrl.state = CTRL_TRF_STATUS_OUT;
				break;

			case CTRL_TRF_DATA_IN:
				//neguje data toggle bit
				ep0_send_data();
				break;

			case CTRL_TRF_STATUS_IN:
				/* koniec no-data control prenosu */
				if(usb_state == USB_ADDR_PENDING) {
					/* nastavi USB adresu zariadenia */
					ADDR = ep0out_setup.bytes.wValue[0];
					if(ADDR != 0)
						/* nova adresa 0 ponecha modul v zakladnom stave */
						usb_state = USB_ADDRESSED;
				}
				/* nasleduje dalsi control transfer */
				ep0_next_trf();
				break;
		}
	} else { /* STAT_OUT */
		/* transakcia prijata */
		if(USB_EP_PID(bdt.ep0out) == PID_SETUP)
			ctrl.state = CTRL_TRF_SETUP;

		switch(ctrl.state) {
			case CTRL_TRF_SETUP: {
				//endian swap
				uint8_t tmp = ep0out_setup.bytes.wLength[0];
				ep0out_setup.bytes.wLength[0] = ep0out_setup.bytes.wLength[1];
				ep0out_setup.bytes.wLength[1] = tmp;
	
				//spracuje setup transakciu 
				ep0_setup();
				break;
			}
			case CTRL_TRF_STATUS_OUT:
				if(bdt.ep0out.bc != 0) {
					//chyba, status faza ma niest 0 bajtov
					EPCTL0_EPSTALL = 1;
					break;
				}
				//koniec control read prenosu, ocakava dalsi control prenos
				ep0_next_trf();
				break;
		}
	}
	CTL_TSUSPEND = 0;
}

/*
Funkcia: ep0_next_trf
Popis: Nastavi konecny automat riadiaceho prenosu a buffer descriptor endpointov
       0 IN a 0 OUT do stavu prijmu setup transakcie.
*/
void ep0_next_trf(void) {
	//ocakava setup transakciu
	ctrl.state = CTRL_TRF_SETUP;
	//IN endpoint vlastni MCU, zacne vysielat DATA0 pid
	bdt.ep0in.u.reg = EPF_OWN_MCU | EPF_DTS | EPF_DATA0;
	//OUT endpoint vlastni SIE, zapne DTS a ocakava DATA0 pid
	bdt.ep0out.bc = sizeof ep0out_setup;
	bdt.ep0out.u.reg = EPF_OWN_SIE | EPF_DTS | EPF_DATA0;
}
//------------------------------------------------------------------------------
//spracovanie ep0out_setup transakcii

/*
Funkcia: ep0_setup
Popis: Spracuje setup transakciu.
*/
static void ep0_setup(void) {
	uint8_t ok = 0; //zatial chyba prikazu
	ctrl.bytes = 0; //a ziadne data na prenos
	ctrl.data = NULL;
	ctrl.trunc = 0;

	if(ep0out_setup.bmRequestType.Type != TYPE_STD)
		//specificke prikazy (class-specific requests) dostane aplikacia
		ok = usb_OnClassReq();
	else {
		//standardny prikaz spracuje tato funkcia
		uint8_t req = ep0out_setup.bytes.bRequest;

		for(;;) {
			if(usb_state >= USB_DEFAULT) {
				/* povolene prikazy: GET_DESCRIPTOR, SET_ADDRESS (mimo
				   CONFIGURED stavu)*/
				if(req == GET_DESCRIPTOR) {
					ok = get_descriptor_hdr();
					break;
#if 0
/*
A device must be able to accept a SetFeature(TEST_MODE, TEST_SELECTOR)
request when in the Default State. Device behavior for other SetFeature requests
while the device is in the Default state is not specified.
*/
				} else if(req == SET_FEATURE) {
					ok = feature_hdr(1/*req == SET_FEATURE*/);
					break;
#endif
				} else if(req == SET_ADDRESS) {
					if(usb_state != USB_CONFIGURED) {
						/* nemozno priradit adresu priamo do ADDR lebo este musi
						   prebehnut status faza so starou adresou (0) */
						usb_state = USB_ADDR_PENDING;	
						ctrl.bytes = 0;
						ok = 1;
					}
					break;
				}
			}

			if(usb_state >= USB_ADDRESSED) {
				/* povolene prikazy: GET/SET_CONFIGURATION, CLEAR/SET_FEATURE,
				   SET_DESCRIPTOR, GET_STATUS */
				if(req == SET_CONFIGURATION) {
					ok = set_configuration_hdr();
					break;
				} else if(req == GET_CONFIGURATION) {
					ok = get_configuracion_hdr();
					break;
				} else if(req == SET_FEATURE || req == CLEAR_FEATURE) {
					ok = feature_hdr(req == SET_FEATURE);
					break;
				} else if(req == GET_STATUS) {
					ok = get_status_hdr();
					break;
				} else if(req == SET_DESCRIPTOR) {
					//nepodporovane
					ok = 0;
					break;
				}
			}

			if(usb_state >= USB_CONFIGURED) {
				/* povolene prikazy: GET/SET_INTERFACE, SYNCHR_FRAME */
				if(req == GET_INTERFACE) {
					ok = get_interface_hdr();
					break;
				} else if(req == SET_INTERFACE) {
					ok = set_interface_hdr();
					break;
				} else if(req == SYNCHR_FRAME) {
					ok = 0; //endpoint 0 nepodporuje isochronne prenosy
					break;
				}
			}
			ok = 0;
			break;
		}
	}
	//reakcia na setup fazu
	if(ok)
		//vykona pozadovany prenos
		ep0_setup_complete();
	else
		//chyba prikazu, posle stall handshake
		EPCTL0_EPSTALL = 1;
}

/*
Funkcia: ep0_setup_complete
Popis: Reakcia na SPRAVNU setup transakciu, vykona pozadovany prenos. Control
       write je nepodporovany.
*/
static void ep0_setup_complete(void) {
	if(ctrl.bytes == 0) {
		//no-data control

		//prejde rovno do stavovej fazy ktora ma format IN DATA1 pid s dlzkou 0B
		ctrl.state = CTRL_TRF_STATUS_IN;
		bdt.ep0in.bc = 0; //0B
		bdt.ep0in.u.reg = EPF_OWN_SIE | EPF_DTS | EPF_DATA1; //IN DATA1 pid
		return;
	}

	if(ctrl.bytes < ep0out_setup.std.wLength &&
		ctrl.bytes % sizeof ep0in_buf == 0)
		ctrl.trunc = 1;

	if(ep0out_setup.bmRequestType.Direction == DIR_DEV_TO_HOST) {
		//control read
		if(ctrl.bytes > ep0out_setup.std.wLength)
			//host ziada menej bajtov ako je typicka dlzka
			ctrl.bytes = ep0out_setup.std.wLength;

		//konecny automat riadiaceho prenosu bude posielat data (IN)
		ctrl.state = CTRL_TRF_DATA_IN;
		//zacina s DATA1 pid, lebo setup bol s DATA0 pid
		bdt.ep0in.u.bits.DATA = 1;

		//posle prvu varku dat
		ep0_send_data();

		//a bez tohoto to proste neide... o.O
		bdt.ep0out.bc = sizeof ep0out_setup;
		bdt.ep0out.u.reg = EPF_OWN_SIE | EPF_DTS | EPF_DATA1;
		return;
	}
	//nepodpruje control write, sem doimplementovat ak je potrebny
	EPCTL0_EPSTALL = 1;
}
//------------------------------------------------------------------------------
//spracovanie standardnych prikazov

/*
Funkcia: get_descriptor_hdr
Navratova hodnota: 0 = spatny prikaz, inak ok
Popis: Spracovanie GET_DESCRIPTOR prikazu.
*/
static uint8_t get_descriptor_hdr(void) {
/* ep0out_setup:
	bRequest	GET_DESCRIPTOR
	wValue		[1] = typ, [0] = index
	wIndex		0 alebo LANG ID
	wLength		pozadovany pocet bajtov na prenesenie, nemenit, nenastavovat! */

	enum {
		DEVICE_DESCRIPTOR = 1,
		CONFIGURATION_DESCRIPTOR = 2,
		STRING_DESCRIPTOR = 3,
		INTERFACE_DESCRIPTOR = 4,
		ENDPOINT_DESCRIPTOR = 5,
		DEVICE_QUALIFIER = 6,
		OTHER_SPEED_DESCRIPTOR = 7,
		INTERFACE_POWER = 8,
		OTG_DESCRIPTOR = 9,
	};

	uint8_t type = ep0out_setup.bytes.wValue[1]; /* typ deskriptoru, enum ^ */
	uint8_t idx = ep0out_setup.bytes.wValue[0]; /* index */
	switch(type) {
		case DEVICE_DESCRIPTOR:
			ctrl.data = (uint8_t*)device_descriptor;
			ctrl.bytes = *ctrl.data; //bLength
			return 1;
		case CONFIGURATION_DESCRIPTOR: {
			union {
				uint16_t u16;
				uint8_t u8[2];
			}len;
			uint8_t tmp;

//			idx--; /* indexovane od 1 lebo index 0 je pre ziadnu konfiguraciu */
			if(idx >= sizeof cfg_desc_array / sizeof cfg_desc_array[0])
				/* index mimo rozsahu */
				return 0;
			ctrl.data = (uint8_t*)cfg_desc_array[idx];
			//wTotalLength v little endian
			len.u16 = *(uint16_t*)(ctrl.data + 2);
			/* endian swap */
			tmp = len.u8[0];
			len.u8[0] = len.u8[1];
			len.u8[1] = tmp;
			ctrl.bytes = len.u16;
			return 1;
		}
		case STRING_DESCRIPTOR:
			if(idx >= sizeof str_desc_array / sizeof str_desc_array[0])
				/* index mimo rozsahu */
				return 0;
			ctrl.data = (uint8_t*)str_desc_array[idx];
			ctrl.bytes = *ctrl.data; /* ukazuje na bLength */
			return 1;
		case INTERFACE_DESCRIPTOR:
		case ENDPOINT_DESCRIPTOR:
			//"not directly accessible"
			return 0;
		case OTHER_SPEED_DESCRIPTOR:
			return 0;
		case INTERFACE_POWER:
			//"obsolete"
			return 0;
		case OTG_DESCRIPTOR:
			//"not directly accessible"
			return 0;
		case DEVICE_QUALIFIER:
			/* full-speed zariadenie nema device qualifier */
			return 0;
	}

	//ziadost o neznamy typ deskriptoru posle aplikacii
	return usb_OnClassReq();
}

/*
Funkcia: set_configuration_hdr
Navratova hodnota: 0 = spatny prikaz, inak ok
Popis: Spracovanie SET_CONFIGURATION prikazu. Aplikacii posiela udalosti
       konfiguracie.
*/
static uint8_t set_configuration_hdr(void) {
/* ep0out_setup:
	bRequest	SET_CONFIGURATION
	wValue		[0] = index
	wIndex		0
	wLength		pozadovany pocet bajtov na prenesenie, nemenit, nenastavovat! */

	uint8_t idx = ep0out_setup.bytes.wValue[0];
	ctrl.data = NULL;
	ctrl.bytes = 0;

	if(idx == 0) {
		//ziadost o deaktivaciu aktivnej konfiguracie
		if(cur_cfg != 0)
			//deaktivuje momentalnu konfiguraciu
			cfg_hdrs[cur_cfg](EVENT_DEACTIVATE);

		usb_state = USB_ADDRESSED;
		return 1;
	}

	//ziadost o aktivaciu nejakej konfiguracie

	idx--; /* indexovane od 1 */
	if(idx >= sizeof cfg_hdrs / sizeof cfg_hdrs[0])
		/* index mimo rozsahu */
		return 0;

	if(cur_cfg != 0)
		//deaktivuje aktivnu konfiguraciu
		cfg_hdrs[cur_cfg - 1](EVENT_DEACTIVATE);
	//aktivuje novu konfiguraciu
	cfg_hdrs[idx](EVENT_ACTIVATE);	
	cur_cfg = idx + 1;

	usb_state = USB_CONFIGURED;
	return 1;
}

/*
Funkcia: get_configuracion_hdr
Navratova hodnota: 0 = spatny prikaz, inak ok
Popis: Posle momentalne aktivnu konfiguraciu ako odpoved na GET_CONFIGURATION
       prikaz.
*/
static uint8_t get_configuracion_hdr(void) {
/* ep0out_setup:
	bRequest	GET_CONFIGURATION
	wValue		0
	wIndex		0
	wLength		1 */

	if(ep0out_setup.bytes.wLength[0] != 1)
		return 0;

	ctrl.data = NULL;
	ctrl.bytes = 1;
	ep0in_buf[0] = cur_cfg; /* vlozi momentalnu konfguraciu */
	return 1;
}

/*
Funkcia: feature_hdr
Parametre: set: 1 = SET_FEATURE, inak CLEAR_FEATURE prikaz
Navratova hodnota: 0 = spatny prikaz, inak ok
Popis: Spracuje SET/CLEAR _FEATURE prikazy.
*/
static uint8_t feature_hdr(uint8_t set) {
/* ep0out_setup:
	bRequest	SET_FEATURE, CLEAR_FEATURE
	wValue		feature selector
	wIndex		0 / interface / endpoint
	wLength		0 */

	uint8_t feature = ep0out_setup.bytes.wValue[0];

	enum {
		ENDPOINT_HALT = 0,
		DEVICE_REMOTE_WAKEUP = 1,
		TEST_MODE = 2,
	};

	if(ep0out_setup.bytes.wLength[0] != 0 || feature >= TEST_MODE)
		return 0;
	if(feature == DEVICE_REMOTE_WAKEUP) {
		if(ep0out_setup.bmRequestType.Recipient != RCPT_DEVICE)
			return 0;
		//zakaze alebo povoli remote wakeup
		if(set)
			usb_status.bits.Remote_wakeup = 1;
		else
			usb_status.bits.Remote_wakeup = 0;
	} else if(feature == ENDPOINT_HALT) {
		//STALL priznak endpointu
		uint8_t *reg = &EPCTL0;
		if(ep0out_setup.bmRequestType.Recipient != RCPT_ENDPOINT)
			return 0;
		if(ep0out_setup.wIndex_ep.Endpoint > 7)
			//taky endpoint na MC9S08JM60 neexistuje
			return 0;
		reg += ep0out_setup.wIndex_ep.Endpoint;
		//EPCTLn_EPSTALL = set
		if(set)
			*reg |= 0x02;
		else
			*reg &= 0xFD;
	}
	return 1;
}

/*
Funkcia: get_status_hdr
Navratova hodnota: 0 = spatny prikaz, inak ok
Popis: Spracuje GET_STATUS prikaz. Posiela remote_wakeup a self_powered
       priznaky.
*/
static uint8_t get_status_hdr(void) {
/* ep0out_setup:
	bRequest	GET_STATUS
	wValue		0
	wIndex		0 / interface / endpoint
	wLength		0 */

	uint8_t recipient = ep0out_setup.bmRequestType.Recipient;

	if(ep0out_setup.bytes.wLength[0] != 2)
		return 0;

	switch(recipient) {
		case RCPT_DEVICE:
			ctrl.data = NULL; //(uint8_t*)&usb_status;
			ep0in_buf[0] = usb_status.bytes[0]; /* remote_wakeup, self_powered */
			ep0in_buf[1] = 0;
			ctrl.bytes = 2; //sizeof usb_status;
			return 1;
		case RCPT_ENDPOINT:
		case RCPT_INTERFACE:
			/* vracaju 00 00 */
			ctrl.data = NULL;
			ctrl.bytes = 2;
			ep0in_buf[0] = 0;
			ep0in_buf[1] = 0;
			return 1;
	}
	return 0;
}

/*
Funkcia: get_interface_hdr
Navratova hodnota: 0 = spatny prikaz, inak ok
Popis: Vracia alternativne nastavenie rozhrania pre zvolene rozhranie ako
       odpoved na GET_INTERFACE prikaz.
*/
static uint8_t get_interface_hdr(void) {
/* ep0out_setup:
	bRequest	GET_INTERFACE
	wValue		0
	wIndex		0 / interface / endpoint
	wLength		0 */

	uint8_t iface = ep0out_setup.bytes.wIndex[0];
	if(ep0out_setup.bytes.wLength[0] != 1)
		return 0;
	if(iface >= sizeof alt_interface_array / sizeof alt_interface_array[0])
		return 0;
	ctrl.data = NULL;
	ctrl.bytes = 1;
	ep0in_buf[0] = alt_interface_array[iface];
}

/*
Funkcia: set_interface_hdr
Navratova hodnota: 0 = spatny prikaz, inak ok
Popis: Nastavi alternativne rozhranie pre zvolene rozhranie ako reakciu na
       SET_INTERFACE prikaz.
*/
static uint8_t set_interface_hdr(void) {
/* ep0out_setup:
	bRequest	SET_INTERFACE
	wValue		0
	wIndex		0 / interface / endpoint
	wLength		0 */

	uint8_t iface = ep0out_setup.bytes.wIndex[0];
	uint8_t alt_iface = ep0out_setup.bytes.wValue[0];
	uint8_t cnt = sizeof alt_interface_array / sizeof alt_interface_array[0];

	if(ep0out_setup.bytes.wLength[0] != 1 || alt_iface >= cnt || iface >= cnt)
		return 0;

	ctrl.data = NULL;
	ctrl.bytes = 0;
	alt_interface_array[iface] = alt_iface;
	return 1;
}
